import logging
from logging import config
from logging.config import dictConfig
import os
from enum import Enum
from typing import NewType

ENV_VAR = "WS_RPC_LOGGING"


class LoggingModes(Enum):
    # don't produce logs
    NO_LOGS = 0
    # Log alongside uvicorn
    UVICORN = 1
    # Simple log calls (no config)
    SIMPLE = 2
    # log via the loguru module
    LOGURU = 3


LoggingMode = NewType('LoggingMode', LoggingModes)


class LoggingConfig:

    def __init__(self) -> None:
        self._mode = None

    config_template = {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "default": {
                "()": "uvicorn.logging.DefaultFormatter",
                "fmt": "%(levelprefix)s %(asctime)s %(message)s",
                "datefmt": "%Y-%m-%d %H:%M:%S",
            },
        },
        "handlers": {
            "default": {
                "formatter": "default",
                "class": "logging.StreamHandler",
            },
        },
        "loggers": {}
    }



    UVICORN_LOGGERS = {
        'uvicorn.error': {
            'propagate': False,
            'handlers': ['default'],
        },
        "fastapi_ws_rpc": {"handlers": ["default"], 'propagate': False, 'level': logging.INFO},
    }

    def get_mode(self):
        # if no one set the mode - set default from ENV or hardcoded default
        if self._mode is None:
            mode = LoggingModes.__members__.get(os.environ.get(ENV_VAR, "").upper(), LoggingModes.SIMPLE)
            self.set_mode(mode)
        return self._mode

    def set_mode(self, mode: LoggingMode = LoggingModes.UVICORN, level= logging.INFO):
        """
        Configure logging. this method calls 'logging.config.dictConfig()' to enable quick setup of logging.
        Call this method before starting the app.
        For more advanced cases use 'logging.config' directly (loggers used by this library are all nested under "fastapi_ws_rpc" logger name)

        Args:
            mode (LoggingMode, optional): The mode to set logging to. Defaults to LoggingModes.UVICORN.
        """
        self._mode = mode
        logging_config = self.config_template.copy()
        # add logs beside uvicorn
        if mode == LoggingModes.UVICORN:
            logging_config["loggers"] = self.UVICORN_LOGGERS.copy()
            logging_config["loggers"]["fastapi_ws_rpc"]["level"] = level
            dictConfig(logging_config)
        elif mode == LoggingModes.SIMPLE:
            pass
        elif mode == LoggingModes.LOGURU:
            pass
        # no logs
        else:
            logging_config["loggers"] = {}
            dictConfig(logging_config)


# Singelton for logging configuration
logging_config = LoggingConfig()


def get_logger(name):
    """
    Get a logger object to log with..
    Called by inner modules for logging. 
    """
    mode = logging_config.get_mode()
    # logging through loguru
    if mode == LoggingModes.LOGURU:
        from loguru import logger
    else:
        # regular python logging
        logger = logging.getLogger(f"fastapi_ws_rpc.{name}")

    return logger



import asyncio
import logging
from typing import Coroutine, Dict, List, Type
from tenacity import retry, wait
import tenacity
from tenacity.retry import retry_if_exception

import websockets
from websockets.exceptions import InvalidStatusCode, WebSocketException, ConnectionClosedError, ConnectionClosedOK

from .rpc_methods import PING_RESPONSE, RpcMethodsBase
from .rpc_channel import RpcChannel, OnConnectCallback, OnDisconnectCallback
from .logger import get_logger
from .simplewebsocket import SimpleWebSocket, JsonSerializingWebSocket

logger = get_logger("RPC_CLIENT")


def isNotInvalidStatusCode(value):
    return not isinstance(value, InvalidStatusCode)

def isNotForbbiden(value)->bool:
    """
    Returns:
        bool: Returns True as long as the given exception value is not InvalidStatusCode with 401 or 403
    """
    return not (isinstance(value, InvalidStatusCode) and (value.status_code == 401 or value.status_code == 403))


class WebSocketRpcClient:
    """
    RPC-client to connect to an WebsocketRPCEndpoint
    Can call methodes exposed by server
    Exposes methods that the server can call
    """

    def logerror(retry_state:tenacity.RetryCallState):
        logger.exception(retry_state.outcome.exception())

    DEFAULT_RETRY_CONFIG = {
        'wait': wait.wait_random_exponential(min=0.1, max=120),
        'retry': retry_if_exception(isNotForbbiden),
        'reraise': True,
        "retry_error_callback":logerror
    }

    # RPC ping check on successful Websocket connection
    # @see wait_on_rpc_ready
    # interval to try RPC-pinging the sever (Seconds)
    WAIT_FOR_INITIAL_CONNECTION = 1
    # How many times to try repinging before rejecting the entire connection
    MAX_CONNECTION_ATTEMPTS = 5

    def __init__(self, uri: str, methods: RpcMethodsBase = None,
                 retry_config=None,
                 default_response_timeout: float = None,
                 on_connect: List[OnConnectCallback] = None,
                 on_disconnect: List[OnDisconnectCallback] = None,
                 keep_alive: float = 0,
                 serializing_socket_cls: Type[SimpleWebSocket] = JsonSerializingWebSocket,
                 **kwargs):
        """
        Args:
            uri (str): server uri to connect to (e.g. 'http://localhost/ws/client1')
            methods (RpcMethodsBase): RPC methods to expose to the server
            retry_config (dict): Tenacity.retry config (@see https://tenacity.readthedocs.io/en/latest/api.html#retry-main-api) 
            default_response_timeout (float): default time in seconds
            on_connect (List[Coroutine]): callbacks on connection being established (each callback is called with the channel)
                                          @note exceptions thrown in on_connect callbacks propagate to the client and will cause connection restart!
            on_disconnect (List[Coroutine]): callbacks on connection termination (each callback is called with the channel)
            keep_alive(float): interval in seconds to send a keep-alive ping, Defaults to 0, which means keep alive is disabled.

            **kwargs: Additional args passed to connect (@see class Connect at websockets/client.py)
                      https://websockets.readthedocs.io/en/stable/api.html#websockets.client.connect


            usage:
                async with  WebSocketRpcClient(uri, RpcUtilityMethods()) as client:
                response = await client.call("echo", {'text': "Hello World!"})
                print (response)
        """
        self.methods = methods or RpcMethodsBase()
        self.connect_kwargs = kwargs
        # Websocket connection
        self.conn = None
        # Websocket object
        self.ws = None
        # URI to connect on
        self.uri = uri
        # Pending requests - id mapped to async-event
        self.requests: Dict[str, asyncio.Event] = {}
        # Received responses
        self.responses = {}
        # Read worker
        self._read_task = None
        # Keep alive (Ping/Pong) task
        self._keep_alive_task = None
        self._keep_alive_interval = keep_alive
        # defaults
        self.default_response_timeout = default_response_timeout
        # RPC channel
        self.channel = None
        self.retry_config = retry_config if retry_config is not None else self.DEFAULT_RETRY_CONFIG
        # Event handlers
        self._on_disconnect = on_disconnect
        self._on_connect = on_connect
        # serialization
        self._serializing_socket_cls = serializing_socket_cls

    async def __connect__(self):
        try:
            # Make sure we don't have any hanging tasks (from previous retry runs)
            self.cancel_tasks()
            logger.info(f"Trying server - {self.uri}")
            # Start connection
            self.conn = websockets.connect(self.uri, **self.connect_kwargs)
            # Get socket and wrap in our serialization class
            raw_ws = await self.conn.__aenter__()
            self.ws = self._serializing_socket_cls(raw_ws)
            # Wrap
            # Init an RPC channel to work on-top of the connection
            self.channel = RpcChannel(self.methods, self.ws, default_response_timeout=self.default_response_timeout)
            # register handlers
            self.channel.register_connect_handler(self._on_connect)
            self.channel.register_disconnect_handler(self._on_disconnect)
            # Start reading incoming RPC calls
            self._read_task = asyncio.create_task(self.reader())
            # start keep alive (if enabled i.e. value isn't 0)
            self._start_keep_alive_task()
            # Wait for RPC channel on the server to be ready (ping check)
            await self.wait_on_rpc_ready()
            # trigger connect handlers
            await self.channel.on_connect()
            return self
        except ConnectionRefusedError:
            logger.info("RPC connection was refused by server")
            raise
        except ConnectionClosedError:
            logger.info("RPC connection lost")
            raise
        except ConnectionClosedOK:
            logger.info("RPC connection closed")
            raise
        except InvalidStatusCode as err:
            logger.info(f"RPC Websocket failed - with invalid status code {err.status_code}")
            raise
        except WebSocketException as err:
            logger.info(f"RPC Websocket failed - with {err}")
            raise
        except OSError as err:
            logger.info("RPC Connection failed - %s", err)
            raise        
        except Exception as err:
            logger.exception("RPC Error")
            raise

    async def __aenter__(self):
        if self.retry_config is False:
            return await self.__connect__()
        else:
            return await retry(**self.retry_config)(self.__connect__)()

    async def __aexit__(self, *args, **kwargs):
        await self.close()
        # close context of underlying socket
        if (hasattr(self.conn, "ws_client")):
            await self.conn.__aexit__(*args, **kwargs)            

    async def close(self):
        logger.info("Closing RPC client")
        # Close underlying connection
        if self.ws is not None:
            await self.ws.close()
        # Notify callbacks (but just once)
        if not self.channel.isClosed():
            # notify handlers (if any)
            await self.channel.on_disconnect()
        # Clear tasks
        self.cancel_tasks()
    
    def cancel_tasks(self):
        # Stop keep alive if enabled
        self._cancel_keep_alive_task()
        # Stop reader - if created
        self.cancel_reader_task()

    def cancel_reader_task(self):
        if self._read_task is not None:
            self._read_task.cancel()
            self._read_task = None

    async def reader(self):
        """
        Read responses from socket worker
        """
        try:
            while True:
                raw_message = await self.ws.recv()
                await self.channel.on_message(raw_message)
        # Graceful external termination options
        # task was canceled
        except asyncio.CancelledError:
            pass
        except websockets.exceptions.ConnectionClosed:
            logger.info("Connection was terminated.")
            await self.close()
        except:
            logger.exception("RPC Reader task failed")
            raise

    async def _keep_alive(self):
        try:
            while True:
                await asyncio.sleep(self._keep_alive_interval)
                answer = await self.ping()
                assert answer.result == PING_RESPONSE
        # Graceful external termination options
        # task was canceled
        except asyncio.CancelledError:
            pass

    async def wait_on_rpc_ready(self):
        received_response = None
        attempt_count = 0
        while received_response is None and attempt_count < self.MAX_CONNECTION_ATTEMPTS:
            try:
                received_response = await asyncio.wait_for(self.ping(), self.WAIT_FOR_INITIAL_CONNECTION)
            except asyncio.exceptions.TimeoutError:
                attempt_count += 1

    async def ping(self):
        logger.debug("Pinging server")
        answer = await self.channel.other._ping_()
        return answer

    def _cancel_keep_alive_task(self):
        if self._keep_alive_task is not None:
            logger.debug("Cancelling keep alive task")
            self._keep_alive_task.cancel()
            self._keep_alive_task = None

    def _start_keep_alive_task(self):
        if self._keep_alive_interval > 0:
            logger.debug(f"Starting keep alive task interval='{self._keep_alive_interval}' seconds")
            self._keep_alive_task = asyncio.create_task(self._keep_alive())

    async def wait_on_reader(self):
        """
        Join on the internal reader task
        """
        try:
            await self._read_task
        except asyncio.CancelledError:
            logger.info(f"RPC Reader task was cancelled.")

    async def call(self, name, args={}, timeout=None):
        """
        Call a method and wait for a response to be received
         Args:
            name (str): name of the method to call on the other side (As defined on the otherside's RpcMethods object)
            args (dict): keyword arguments to be passeds to otherside method
        """
        return await self.channel.call(name, args, timeout=timeout)

    @property
    def other(self):
        """
        Proxy object to call methods on the other side
        """
        return self.channel.other



import asyncio
from typing import Coroutine, List, Type
from fastapi import WebSocket, WebSocketDisconnect

from .connection_manager import ConnectionManager
from .rpc_channel import RpcChannel
from .rpc_methods import RpcMethodsBase
from .logger import get_logger
from .schemas import WebSocketFrameType
from .simplewebsocket import SimpleWebSocket, JsonSerializingWebSocket

logger = get_logger("RPC_ENDPOINT")

class WebSocketSimplifier(SimpleWebSocket):
    """
    Simple warpper over FastAPI WebSocket to ensure unified interface for send/recv
    """

    def __init__(self, websocket: WebSocket, frame_type: WebSocketFrameType = WebSocketFrameType.Text):
        self.websocket = websocket
        self.frame_type = frame_type

    @property
    def send(self):
        if self.frame_type == WebSocketFrameType.Binary:
            return self.websocket.send_bytes
        else:
            return self.websocket.send_text


    @property
    def recv(self):
        if self.frame_type == WebSocketFrameType.Binary:
            return self.websocket.receive_bytes
        else:
            return self.websocket.receive_text

    async def close(self, code: int = 1000):
        return await self.websocket.close(code)

class WebsocketRPCEndpoint:
    """
    A websocket RPC sever endpoint, exposing RPC methods
    """

    def __init__(self, methods: RpcMethodsBase = None,
                 manager: ConnectionManager = None,
                 on_disconnect: List[Coroutine] = None,
                 on_connect: List[Coroutine] = None,
                 frame_type: WebSocketFrameType = WebSocketFrameType.Text,
                 serializing_socket_cls: Type[SimpleWebSocket] = JsonSerializingWebSocket,
                 rpc_channel_get_remote_id: bool = False):
        """[summary]

        Args:
            methods (RpcMethodsBase): RPC methods to expose
            manager ([ConnectionManager], optional): Connection tracking object. Defaults to None (i.e. new ConnectionManager()).
            on_disconnect (List[coroutine], optional): Callbacks per disconnection
            on_connect(List[coroutine], optional): Callbacks per connection (Server spins the callbacks as a new task, not waiting on it.)
        """
        self.manager = manager if manager is not None else ConnectionManager()
        self.methods = methods if methods is not None else RpcMethodsBase()
        # Event handlers
        self._on_disconnect = on_disconnect
        self._on_connect = on_connect
        self._frame_type = frame_type
        self._serializing_socket_cls = serializing_socket_cls
        self._rpc_channel_get_remote_id = rpc_channel_get_remote_id


    async def main_loop(self, websocket: WebSocket, client_id: str = None, **kwargs):
        try:
            await self.manager.connect(websocket)
            logger.info(f"Client connected", {'remote_address':websocket.client})
            simple_websocket = self._serializing_socket_cls(WebSocketSimplifier(websocket, frame_type=self._frame_type))
            channel = RpcChannel(self.methods, simple_websocket, sync_channel_id=self._rpc_channel_get_remote_id, **kwargs)
            # register connect / disconnect handler
            channel.register_connect_handler(self._on_connect)
            channel.register_disconnect_handler(self._on_disconnect)
            # trigger connect handlers
            await channel.on_connect()
            try:
                while True:
                    data = await simple_websocket.recv()
                    await channel.on_message(data)
            except WebSocketDisconnect:
                logger.info(f"Client disconnected - {websocket.client.port} :: {channel.id}")
                await self.handle_disconnect(websocket, channel)
            except:
                # cover cases like - RuntimeError('Cannot call "send" once a close message has been sent.')
                logger.info(f"Client connection failed - {websocket.client.port} :: {channel.id}")
                await self.handle_disconnect(websocket, channel)
        except:
            logger.exception(f"Failed to serve - {websocket.client.port}")
            self.manager.disconnect(websocket)

    async def handle_disconnect(self, websocket, channel):
        self.manager.disconnect(websocket)
        await channel.on_disconnect()


    async def on_connect(self, channel, websocket):
        """
        Called upon new client connection
        """
        # Trigger connect callback if available
        if (self._on_connect is not None):
            asyncio.create_task(self._on_connect(channel, websocket))

    def register_route(self, router, path="/ws"):
        """
        Register this endpoint as a default websocket route on the given router
        Args:
            router: FastAPI router to load route onto
            path (str, optional): the route path. Defaults to "/ws".
        """

        @router.websocket(path)
        async def websocket_endpoint(websocket: WebSocket):
            await self.main_loop(websocket)



"""
Definition for an RPC channel protocol on top of a websocket - enabling bi-directional request/response interactions
"""
import asyncio
from inspect import _empty, getmembers, ismethod, signature
from typing import Any, Coroutine, Dict, List
from asyncio.coroutines import coroutine

from pydantic import ValidationError

from .utils import gen_uid
from .rpc_methods import EXPOSED_BUILT_IN_METHODS, NoResponse, RpcMethodsBase
from .schemas import RpcMessage, RpcRequest, RpcResponse

from .logger import get_logger
logger = get_logger("RPC_CHANNEL")


class DEAFULT_TIMEOUT:
    pass

class RemoteValueError(ValueError):
    pass

class RpcException(Exception):
    pass

class RpcChannelClosedException(Exception):
    """
    Raised when the channel is closed mid-operation
    """
    pass

class UnknownMethodException(RpcException):
    pass


class RpcPromise:
    """
    Simple Event and id wrapper/proxy
    Holds the state of a pending request
    """

    def __init__(self, request: RpcRequest):
        self._request = request
        self._id = request.call_id
        # event used to wait for the completion of the request (upon receiving its matching response)
        self._event = asyncio.Event()

    @property
    def request(self):
        return self._request

    @property
    def call_id(self):
        return self._id

    def set(self):
        """
        Signal compeltion of request with received response
        """
        self._event.set()

    def wait(self):
        """
        Wait on the internal event - triggered on response
        """
        return self._event.wait()


class RpcProxy:
    """
    Helper class
    provide a __call__ interface for an RPC method over a given channel
    """

    def __init__(self, channel, method_name) -> None:
        self.method_name = method_name
        self.channel = channel

    def __call__(self, **kwds: Any) -> Any:
        return self.channel.call(self.method_name, args=kwds)


class RpcCaller:
    """
    Helper class provide an object (aka other) with callable methods for each remote method on the otherside
    """

    def __init__(self, channel, methods=None) -> None:
        self._channel = channel
        self._method_names = [method[0] for method in getmembers(
            methods, lambda i: ismethod(i))] if methods is not None else None


    def __getattribute__(self, name: str):
        if (not name.startswith("_") or name in EXPOSED_BUILT_IN_METHODS) and (self._method_names is None or name in self._method_names):
            return RpcProxy(self._channel, name)
        else:
            return super().__getattribute__(name)


# Callback signatures
async def OnConnectCallback(channel):
    pass

async def OnDisconnectCallback(channel):
    pass

async def OnErrorCallback(channel, err:Exception):
    pass


class RpcChannel:
    """
    A wire agnostic json-rpc channel protocol for both server and client.
    Enable each side to send RPC-requests (calling exposed methods on other side) and receive rpc-responses with the return value

    provides a .other property for callign remote methods.
    e.g. answer = channel.other.add(a=1,b=1) will (For example) ask the other side to perform 1+1 and will return an RPC-response of 2
    """

    def __init__(self, methods: RpcMethodsBase, socket, channel_id=None, default_response_timeout=None, sync_channel_id=False, **kwargs):
        """

        Args:
            methods (RpcMethodsBase): RPC methods to expose to other side
            socket: socket object providing simple send/recv methods
            channel_id (str, optional): uuid for channel. Defaults to None in which case a random UUID is generated.
            default_response_timeout(int, optional) default timeout for RPC call responses. Defaults to None - i.e. no timeout
            sync_channel_id(bool, optional) should get the other side of the channel id, helps to identify connections, cost a bit networking time.
                Defaults to False - i.e. not getting the other side channel id
        """
        self.methods = methods._copy_()
        # allow methods to access channel (for recursive calls - e.g. call me as a response for me calling you)
        self.methods._set_channel_(self)
        # Pending requests - id-mapped to async-event
        self.requests: Dict[str, asyncio.Event] = {}
        # Received responses
        self.responses = {}
        self.socket = socket
        # timeout
        self.default_response_timeout = default_response_timeout
        # Unique channel id
        self.id = channel_id if channel_id is not None else gen_uid()
        # flag control should we retrieve the channel id of the other side
        self._sync_channel_id = sync_channel_id
        # The channel id of the other side (if sync_channel_id is True)
        self._other_channel_id = None
        # asyncio event to check if we got the channel id of the other side
        self._channel_id_synced = asyncio.Event()
        #
        # convineice caller
        # TODO - pass remote methods object to support validation before call
        self.other = RpcCaller(self)
        # core event callback regsiters
        self._connect_handlers:List[OnConnectCallback] = []
        self._disconnect_handlers:List[OnDisconnectCallback] = []
        self._error_handlers = []
        # internal event
        self._closed = asyncio.Event()

        # any other kwarg goes straight to channel context (Accessible to methods)
        self._context = kwargs or {}

    @property
    def context(self) -> Dict[str, Any]:
        return self._context

    async def get_other_channel_id(self) -> str:
        """
        Method to get the channel id of the other side of the channel
        The _channel_id_synced verify we have it
        Timeout exception can be raised if the value isn't available
        """
        await asyncio.wait_for(self._channel_id_synced.wait(), self.default_response_timeout)
        return self._other_channel_id

    def get_return_type(self, method):
        method_signature = signature(method)
        return method_signature.return_annotation if method_signature.return_annotation is not _empty else str

    async def send(self, data):
        """
        For internal use. wrap calls to underlying socket
        """
        await self.socket.send(data)

    async def receive(self):
        """
        For internal use. wrap calls to underlying socket
        """
        return await self.socket.recv()

    async def close(self):
        res = await self.socket.close()
        # signal closer
        self._closed.set()
        return res

    def isClosed(self):
        return self._closed.is_set()

    async def wait_until_closed(self):
        return await self._closed.wait()

    async def on_message(self, data):
        """
        Handle an incoming RPC message
        This is the main function servers/clients using the channel need to call (upon reading a message on the wire)
        """
        try:
            message = RpcMessage.parse_obj(data)
            if message.request is not None:
                await self.on_request(message.request)
            if message.response is not None:
                await self.on_response(message.response)
        except ValidationError as e:
            logger.error(f"Failed to parse message", {'message':data, 'error':e})
            await self.on_error(e)
        except Exception as e:
            await self.on_error(e)
            raise


    def register_connect_handler(self, coros:List[OnConnectCallback]=None):
        """
        Register a connection handler callback that will be called (As an async task)) with the channel
        Args:
            coros (List[Coroutine]): async callback
        """
        if coros is not None:
            self._connect_handlers.extend(coros)

    def register_disconnect_handler(self, coros:List[OnDisconnectCallback]=None):
        """
        Register a disconnect handler callback that will be called (As an async task)) with the channel id
        Args:
            coros (List[Coroutine]): async callback
        """
        if coros is not None:
            self._disconnect_handlers.extend(coros)

    def register_error_handler(self, coros:List[OnErrorCallback]=None):
        """
        Register an error handler callback that will be called (As an async task)) with the channel and triggered error.
        Args:
            coros (List[Coroutine]): async callback
        """
        if coros is not None:
            self._error_handlers.extend(coros)

    async def on_handler_event(self, handlers, *args, **kwargs):
        await asyncio.gather(*(callback(*args, **kwargs) for callback in handlers))

    async def on_connect(self):
        """
        Run get other channel id if sync_channel_id is True
        Run all callbacks from self._connect_handlers
        """
        if self._sync_channel_id:
            self._get_other_channel_id_task = asyncio.create_task(self._get_other_channel_id())
        await self.on_handler_event(self._connect_handlers, self)

    async def _get_other_channel_id(self):
        """
        Perform call to the other side of the channel to get its channel id
        Each side is generating the channel id by itself so there is no way to identify a connection without this sync
        """
        if self._other_channel_id is None:
            other_channel_id = await self.other._get_channel_id_()
            self._other_channel_id = other_channel_id.result if other_channel_id and other_channel_id.result else None
            if self._other_channel_id is None:
                raise RemoteValueError()
            # update asyncio event that we received remote channel id
            self._channel_id_synced.set()
            return self._other_channel_id
        else:
            return self._other_channel_id

    async def on_disconnect(self):
        # disconnect happend - mark the channel as closed
        self._closed.set()
        await self.on_handler_event(self._disconnect_handlers, self)

    async def on_error(self, error:Exception):
        await self.on_handler_event(self._error_handlers, self, error)

    async def on_request(self, message: RpcRequest):
        """
        Handle incoming RPC requests - calling relevant exposed method
        Note: methods prefixed with "_" are protected and ignored.

        Args:
            message (RpcRequest): the RPC request with the method to call
        """
        # TODO add exception support (catch exceptions and pass to other side as response with errors)
        logger.debug("Handling RPC request - %s", {'request':message, 'channel':self.id})
        method_name = message.method
        # Ignore "_" prefixed methods (except the built in "_ping_")
        if (isinstance(method_name, str) and (not method_name.startswith("_") or method_name in EXPOSED_BUILT_IN_METHODS)):
            method = getattr(self.methods, method_name)
            if callable(method):
                result = await method(**message.arguments)
                if result is not NoResponse:
                    # get indicated type
                    result_type = self.get_return_type(method)
                    # if no type given - try to convert to string
                    if result_type is str and type(result) is not str:
                        result = str(result)
                    response = RpcMessage(response=RpcResponse[result_type](
                        call_id=message.call_id, result=result, result_type=getattr(result_type, "__name__", getattr(result_type, "_name", "unknown-type"))))
                    await self.send(response)

    def get_saved_promise(self, call_id):
        return self.requests[call_id]

    def get_saved_response(self, call_id):
        return self.responses[call_id]

    def clear_saved_call(self, call_id):
        del self.requests[call_id]
        del self.responses[call_id]

    async def on_response(self, response: RpcResponse):
        """
        Handle an incoming response to a previous RPC call

        Args:
            response (RpcResponse): the received response
        """
        logger.debug("Handling RPC response - %s", {'response':response})
        if response.call_id is not None and response.call_id in self.requests:
            self.responses[response.call_id] = response
            promise = self.requests[response.call_id]
            promise.set()

    async def wait_for_response(self, promise, timeout=DEAFULT_TIMEOUT) -> RpcResponse:
        """
        Wait on a previously made call
        Args:
            promise (RpcPromise): the awaitable-wrapper returned from the RPC request call
            timeout (int, None, or DEAFULT_TIMEOUT): the timeout to wait on the response, defaults to DEAFULT_TIMEOUT.
                - DEAFULT_TIMEOUT - use the value passed as 'default_response_timeout' in channel init
                - None - no timeout
                - a number - seconds to wait before timing out
        Raises:
            asyncio.exceptions.TimeoutError - on timeout
            RpcChannelClosedException - if the channel fails before wait could be completed
        """
        if timeout is DEAFULT_TIMEOUT:
            timeout = self.default_response_timeout
        # wait for the promise or until the channel is terminated
        for coro in asyncio.as_completed([promise.wait(), self._closed.wait()], timeout=timeout):
            await coro
            break
        response = self.responses.get(promise.call_id, NoResponse)
        # if the channel was closed before we could finish
        if response is NoResponse:
            raise RpcChannelClosedException(f"Channel Closed before RPC response for {promise.call_id} could be received")
        self.clear_saved_call(promise.call_id)
        return response

    async def async_call(self, name, args={}, call_id=None) -> RpcPromise:
        """
        Call a method and return the event and the sent message (including the chosen call_id)
        use self.wait_for_response on the event and call_id to get the return value of the call
        Args:
            name (str): name of the method to call on the other side
            args (dict): keyword args to pass tot he other side
            call_id (string, optional): a UUID to use to track the call () - override only with true UUIDs
        """
        call_id = call_id or gen_uid()
        msg = RpcMessage(request=RpcRequest(
            method=name, arguments=args, call_id=call_id))
        logger.debug("Calling RPC method - %s", {'message':msg})
        await self.send(msg)
        promise = self.requests[msg.request.call_id] = RpcPromise(msg.request)
        return promise

    async def call(self, name, args={}, timeout=DEAFULT_TIMEOUT):
        """
        Call a method and wait for a response to be received
        """
        promise = await self.async_call(name, args)
        return await self.wait_for_response(promise, timeout=timeout)







from typing import Dict, Generic, List, Optional, TypeVar
from enum import Enum

from pydantic import BaseModel
from pydantic.generics import GenericModel

UUID = str


class RpcRequest(BaseModel):
    method: str
    arguments: Optional[Dict] = {}
    call_id: Optional[UUID] = None


ResponseT = TypeVar('ResponseT')


class RpcResponse(GenericModel, Generic[ResponseT]):
    result: ResponseT
    result_type: Optional[str]
    call_id: Optional[UUID] = None


class RpcMessage(BaseModel):
    request: Optional[RpcRequest] = None
    response: Optional[RpcResponse] = None


class WebSocketFrameType(str, Enum):
    Text = "text"
    Binary = "binary"


from typing import Any
from abc import ABC, abstractmethod
import json

class SimpleWebSocket(ABC):
    """
    Abstract base class for all websocket related wrappers.
    """
    @abstractmethod
    def send(self, msg):
        pass

    @abstractmethod
    def recv(self):
        pass

    @abstractmethod
    def close(self, code: int = 1000):
        pass


class JsonSerializingWebSocket(SimpleWebSocket):
    def __init__(self, websocket: SimpleWebSocket):
        self._websocket = websocket

    def _serialize(self, msg):
        return msg.json()

    def _deserialize(self, buffer):
        return json.loads(buffer)

    async def send(self, msg):
        await self._websocket.send(self._serialize(msg))

    async def recv(self):
        msg = await self._websocket.recv()

        return self._deserialize(msg)

    async def close(self, code: int = 1000):
        await self._websocket.close(code)




from typing import List

from fastapi import WebSocket


class ConnectionManager:
    def __init__(self):
        self.active_connections: List[WebSocket] = []

    async def connect(self, websocket: WebSocket):
        await websocket.accept()
        self.active_connections.append(websocket)

    def disconnect(self, websocket: WebSocket):
        self.active_connections.remove(websocket)





import asyncio
import os
import sys
import typing
import copy

from fastapi import FastAPI, WebSocket, WebSocketDisconnect
from pydantic import BaseModel

from .connection_manager import ConnectionManager
from .schemas import RpcRequest, RpcResponse
from .utils import gen_uid

PING_RESPONSE = "pong"
# list of internal methods that can be called from remote
EXPOSED_BUILT_IN_METHODS =  ['_ping_', '_get_channel_id_']
# NULL default value - indicating no response was received
class NoResponse:
    pass


class RpcMethodsBase:
    """
    The basic interface RPC channels excpets method groups to implement.
     - create copy of the method object
     - set channel
     - provide '_ping_' for keep-alive
    """

    def __init__(self):
        self._channel = None

    def _set_channel_(self, channel):
        """
        Allows the channel to share access to its functions to the methods once nested under it
        """
        self._channel = channel

    @property
    def channel(self):
        return self._channel

    def _copy_(self):
        """ Simple copy ctor - overriding classes may need to override copy as well."""
        return copy.copy(self)

    async def _ping_(self)->str:
        """
        built in ping for keep-alive
        """
        return PING_RESPONSE

    async def _get_channel_id_(self)->str:
        """
        built in channel id to better identify your remote
        """
        return self._channel.id


class ProcessDetails(BaseModel):
    pid: int = os.getpid()
    cmd: typing.List[str] = sys.argv
    workingdir: str = os.getcwd()

class RpcUtilityMethods(RpcMethodsBase):
    """
    A simple set of RPC functions useful for management and testing
    """

    def __init__(self):
        """
        endpoint (WebsocketRPCEndpoint): the endpoint these methods are loaded into
        """
        super().__init__()

    async def get_proccess_details(self) -> ProcessDetails:
        return ProcessDetails()

    async def call_me_back(self, method_name="", args={}) -> str:
        if self.channel is not None:
            # generate a uid we can use to track this request
            call_id = gen_uid()
            # Call async -  without waiting to avoid locking the event_loop
            asyncio.create_task(self.channel.async_call(method_name, args=args, call_id=call_id))
            # return the id- which can be used to check the response once it's received
            return call_id

    async def get_response(self, call_id="") -> typing.Any:
        if self.channel is not None:
            res = self.channel.get_saved_response(call_id)
            self.channel.clear_saved_call(call_id)
            return res

    async def echo(self, text: str) -> str:
        return text



import datetime
import os
import re
import uuid
from datetime import timedelta
from random import SystemRandom, randrange

__author__ = 'OrW'


class RandomUtils(object):
    @staticmethod
    def gen_cookie_id():
        return os.urandom(16).encode('hex')

    @staticmethod
    def gen_uid():
        return uuid.uuid4().hex

    @staticmethod
    def gen_token(size=256):
        if size % 2 != 0:
            raise ValueError("Size in bits must be an even number.")
        return uuid.UUID(int=SystemRandom().getrandbits(size/2)).hex + \
            uuid.UUID(int=SystemRandom().getrandbits(size/2)).hex

    @staticmethod
    def random_datetime(start=None, end=None):
        """
        This function will return a random datetime between two datetime
        objects.
        If no range is provided - a last 24-hours range will be used
        :param datetime,None start: start date for range, now if None
        :param datetime,None end: end date for range, next 24-hours if None
        """
        # build range
        now = datetime.datetime.now()
        start = start or now
        end = end or now + timedelta(hours=24)
        delta = end - start
        int_delta = (delta.days * 24 * 60 * 60) + delta.seconds
        # Random time
        random_second = randrange(int_delta)
        # Return random date
        return start + timedelta(seconds=random_second)


gen_uid = RandomUtils.gen_uid
gen_token = RandomUtils.gen_token


class StringUtils(object):
    @staticmethod
    def convert_camelcase_to_underscore(name, lower=True):
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        res = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1)
        if lower:
            return res.lower()
        else:
            return res.upper()



from setuptools import setup, find_packages


def get_requirements(env=""):
    if env:
        env = "-{}".format(env)
    with open("requirements{}.txt".format(env)) as fp:
        return [x.strip() for x in fp.read().split("\n") if not x.startswith("#")]


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='fastapi_websocket_rpc',
    version='0.1.21',
    author='Or Weis',
    author_email="or@authorizon.com",
    description="A fast and durable bidirectional JSON RPC channel over Websockets and FastApi.",
    long_description_content_type="text/markdown",
    long_description=long_description,
    url="https://github.com/authorizon/fastapi_websocket_rpc",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Internet :: WWW/HTTP :: WSGI"
    ],
    python_requires='>=3.7',
    install_requires=get_requirements(),
)



from .rpc_methods import RpcMethodsBase, RpcUtilityMethods
from .websocket_rpc_client import WebSocketRpcClient
from .websocket_rpc_endpoint import WebsocketRPCEndpoint
from .rpc_channel import RpcChannel
from .schemas import WebSocketFrameType